import ctypes
from ctypes import*
import os
import math  

# Definitions
STRING = c_char_p
DWORD = c_ulong
ULONG = c_ulong
WORD = c_ushort
USHORT = c_ushort
SHORT = c_short
UCHAR = c_ubyte
BYTE = c_byte
UINT = c_uint
LPBYTE = POINTER(c_ubyte)
CHAR = c_char
LPBOOL = POINTER(c_int)
PUCHAR = POINTER(c_ubyte)
PCHAR = STRING
PVOID = c_void_p
INT = c_int
LPTSTR = STRING
LPDWORD = POINTER(DWORD)
LPWORD = POINTER(WORD)
PULONG = POINTER(ULONG)
LPVOID = PVOID
VOID = None
ULONGLONG = c_ulonglong
HANDLE = c_void_p
BOOL = c_bool


#Handle and status
FT_STATUS = ULONG
FT_HANDLE = PVOID
FT4222_STATUS = ULONG

ftstatus = FT_STATUS(0)
numOfDevices = ULONG(0)

#Load liberies FT4222 and ftd2xx
ftd2xx = windll.LoadLibrary(os.path.dirname(os.path.realpath(__file__)) + "/ftd2xx.dll")
FT4222 = cdll.LoadLibrary(os.path.dirname(os.path.realpath(__file__)) + "/LibFT4222.dll")

#Define functions of ftd2xxp
ftd2xx.FT_CreateDeviceInfoList.restype = FT_STATUS
ftd2xx.FT_CreateDeviceInfoList.argtypes = [LPDWORD]

#FT_CreateDeviceInfoList
ftd2xx.FT_GetDeviceInfoDetail.restype = FT_STATUS
ftd2xx.FT_GetDeviceInfoDetail.argtypes = [DWORD, LPDWORD, LPDWORD, LPDWORD, LPDWORD, LPVOID, LPVOID, LPVOID]

#FT_OpenEx
ftd2xx.FT_OpenEx.restype = FT_STATUS
ftd2xx.FT_OpenEx.argtypes = [STRING, INT, HANDLE]

#FT_SetLatencyTimer
ftd2xx.FT_SetLatencyTimer.restype = FT_STATUS
ftd2xx.FT_SetLatencyTimer.argtypes = [HANDLE, BYTE]

#FT_SetLatencyTimer
ftd2xx.FT_SetUSBParameters.restype = FT_STATUS
ftd2xx.FT_SetUSBParameters.argtypes = [HANDLE, ULONG, ULONG]

#FT_Close
ftd2xx.FT_Close.restype = FT_STATUS
ftd2xx.FT_Close.argtypes = [HANDLE]

################################################################
#FT4222_SetClock
FT4222.FT4222_SetClock.restype = FT4222_STATUS
FT4222.FT4222_SetClock.argtypes = [HANDLE, INT]

#FT4222_SPIMaster_Init
FT4222.FT4222_SPIMaster_Init.restype = FT4222_STATUS
FT4222.FT4222_SPIMaster_Init.argtypes = [HANDLE, INT, INT, INT, INT, BYTE]

#FT4222_SPIMaster_SingleReadWrite
FT4222.FT4222_SPIMaster_SingleReadWrite.restype = FT4222_STATUS

######## 
#Convert returned byte array to a decimal value
def bytes2decimal(Array):
    if len(Array) == 3:
        decimal = Array[1]*256 + Array[2]
        return decimal
    else:
        print("Wrong array length")


# Read a register value 
def ReadRegister(Handle, RegisterValue):
    #Appending address to fit with DISB format. 
    AppendedRegAddres = RegisterValue * 4 + 2

    #Initializing different inputs for ReadWrite function
    Length = ctypes.c_ushort(3)
    TransferSize = ctypes.c_ushort()
    TRUE = ctypes.c_bool(1)
    readbuf = (c_ubyte*3)()
    writebuf = (c_ubyte*3)()
    #Create a writebuffer, requesting data read from chosen address
    writebuf[0] = AppendedRegAddres ; writebuf[1] = 0 ; writebuf[2] = 0 
    #Read DATA
    ft4222Status = FT4222.FT4222_SPIMaster_SingleReadWrite(ftHandleOutCom, ctypes.byref(readbuf), ctypes.byref(writebuf), Length, ctypes.byref(TransferSize), TRUE)
    #Convert the 2 byte value to single decimal value if Read succesful. Otherwise print error message
    if ft4222Status == 0:
        DecimalValue = bytes2decimal(readbuf)
        return DecimalValue
    else:
        print("Something went wrong during the transfer process", ft4222Status)

# Set a new register value
def SetRegister(Handle, RegisterValue, NewRegValue):
    #Appending address to fit with DISB format. 
    AppendedRegAddres = RegisterValue * 4
    
    #Initializing different inputs for ReadWrite function
    Length = ctypes.c_ushort(3)
    TransferSize = ctypes.c_ushort()
    TRUE = ctypes.c_bool(1)
    readbuf = (c_ubyte*3)()
    writebuf = (c_ubyte*3)()
    #Create a writebuffer, choosing the correct register and transferring data
    writebuf[0] = AppendedRegAddres ; writebuf[1] = 0 ; writebuf[2] = 0 
    #If the new value cannot be contained within a single uByte, split it into two bytes.
    if NewRegValue > 255:
        writebuf[1] = math.floor(NewRegValue/256); writebuf[2] = NewRegValue - 256 * math.floor(NewRegValue/255); 
    else:
        writebuf[1] = 0; writebuf[2] = NewRegValue
    
    #Write the new register value to the chosen register.
    ft4222Status = FT4222.FT4222_SPIMaster_SingleReadWrite(ftHandleOutCom, ctypes.byref(readbuf), ctypes.byref(writebuf), Length, ctypes.byref(TransferSize), TRUE)
    #Check if the register was succesfully set.
    if ft4222Status != 0:
        print("Something went wrong during the transfer process", ft4222Status)

#Get a spectrum from the spectrometer
def GetSpectrum(HandleREG, HandleStream):

    #Perform soft reset of image buffer
    SetRegister(HandleREG, 8, 16)
    #Trigger an exposure
    SetRegister(HandleREG, 8, 1)

    #Readout value of first and last pixel to determine framebuffer length
    FirstPixel  = ReadRegister(HandleREG, 23)
    LastPixel   = ReadRegister(HandleREG, 24)

    #Initializing different inputs for ReadWrite function
    Datalength = (LastPixel - FirstPixel + 1) * 2
    readbuf = (c_ubyte*Datalength)()
    writebuf = (c_ubyte*Datalength)()
    Length = ctypes.c_ushort(Datalength)
    TransferSize = ctypes.c_ushort()
    TRUE = ctypes.c_bool(1)
    #Create an empty array for the spectrum to be written to. 
    Spectrum = []

    #Check if the framebuffer is full and spectrum is ready to be read.
    NumberOfPixReady = 0
    while (LastPixel - FirstPixel ) > NumberOfPixReady:
        NumberOfPixReady = ReadRegister(HandleREG, 12)
       
    #Read spectrum
    ft4222Status = FT4222.FT4222_SPIMaster_SingleReadWrite(HandleStream, ctypes.byref(readbuf), ctypes.byref(writebuf), Datalength, ctypes.byref(TransferSize), TRUE)

    #If succesful combine the ubyte values to 16 bit values (65535). Else display error message.  
    if ft4222Status == 0:
        for i in range(0, Datalength, 2):
            Spectrum.append(readbuf[i] * 256 + readbuf[i + 1])
        return Spectrum
    else:
        print("Something went wrong during the transfer process", ft4222Status)


#Device class
def list_devices():
    numOfDevices = ctypes.c_ulong()
    ftd2xx.FT_CreateDeviceInfoList(ctypes.byref(numOfDevices))

    flags = ctypes.c_ulong()
    typ = ctypes.c_ulong()
    ID = ctypes.c_ulong()
    LocID = ctypes.c_ulong()
    serial = ctypes.create_string_buffer(20)
    desc = ctypes.create_string_buffer(64)
    handle = ctypes.c_void_p()

    ret = []
    for i in range(numOfDevices.value):
        ftStatus = ftd2xx.FT_GetDeviceInfoDetail(i, ctypes.byref(flags), ctypes.byref(typ),
                                             ctypes.byref(ID), ctypes.byref(LocID), serial,
                                             desc, ctypes.byref(handle))
        
        #if ftstatus != FT_OK :
         #   print("why")
        ret.append((i, flags.value, typ.value,
                       ID.value, LocID.value, 
                       serial.value.decode("utf-8") , desc.value.decode("utf-8") ))

    return ret



################################################################################################################################
################################################################################################################################
#Start of program.
################################################################################################################################
################################################################################################################################
#Check for available device and collect information on them.
ret = list_devices()

#Set up 3 handles used for register R/W, bulk reads and GPIO
ftHandleOutStream = ctypes.c_void_p()
ftHandleOutCom = ctypes.c_void_p()
ftHandleOutGPIO = ctypes.c_void_p()

SerialNumberA = STRING("A".encode('utf-8'))
SerialNumberB = STRING("B".encode('utf-8'))
SerialNumberC = STRING("C".encode('utf-8'))
#Open 3 devices with their associated handle based on serialnumber.
OPEN_BY_SERIAL = INT(1)
ftStatus = ftd2xx.FT_OpenEx(SerialNumberA, OPEN_BY_SERIAL, ctypes.byref(ftHandleOutStream))
print("A ", ftStatus)
ftStatus = ftd2xx.FT_OpenEx(SerialNumberB, OPEN_BY_SERIAL, ctypes.byref(ftHandleOutCom))
print("B ",ftStatus)
ftStatus = ftd2xx.FT_OpenEx(SerialNumberC, OPEN_BY_SERIAL, ctypes.byref(ftHandleOutGPIO))
print("C ", ftStatus)

###########
#Set latency and USB parameter for data transfer
UcLatency = BYTE(20);
ftStatus = ftd2xx.FT_SetLatencyTimer(ftHandleOutStream, UcLatency)
print("FT_SetLatencyTimer ", ftStatus)

OutTransferSize = ULONG(4096);
InTransferSize = ULONG(65536);
ftStatus = ftd2xx.FT_SetUSBParameters(ftHandleOutStream, InTransferSize, OutTransferSize)
print("FT_SetUSBParameters ", ftStatus)


##################
#Init MASTER SPI
##################
#Initialize a SPI connection with the R/W and bulk read handles.
ft4222Status = FT4222.FT4222_SetClock(ftHandleOutStream, 0) 


ft4222Status = FT4222.FT4222_SPIMaster_Init(ftHandleOutCom, 1, 2, 0, 1, 2);
ft4222Status = FT4222.FT4222_SPIMaster_Init(ftHandleOutStream, 1, 2, 0, 1, 2);   

##################
#READ FROM DISB REGISTER
##################
#Read different register values using the ReadRegister function
PCB_SN = ReadRegister(ftHandleOutCom, 1)
print("\nPCB SERIAL NUMBER ", PCB_SN)

HardwareVersion = ReadRegister(ftHandleOutCom, 2)
print("Hardware Version ", HardwareVersion)

FirmwareVersion = ReadRegister(ftHandleOutCom, 3)
print("Firmware Version ", FirmwareVersion)

DetectorType = ReadRegister(ftHandleOutCom, 4)
print("Detector Type ", DetectorType)

PixelPerImage = ReadRegister(ftHandleOutCom, 5)
print("Pixels per image ", PixelPerImage)

NumberofCaliChars = ReadRegister(ftHandleOutCom, 6)
print("Number of Calibration Chars ", NumberofCaliChars)

print("Wavelength Calibration values")
for i in range((int)(NumberofCaliChars/14)):
    combinedCaliChar = []
    for j in range(14):
        caliChar = ReadRegister(ftHandleOutCom, 7) 
        combinedCaliChar.append(chr(caliChar))
    CalibrationValue = ''.join(map(str,combinedCaliChar))
    print(CalibrationValue)


##################
#SET DISB REGISTER
##################
#Set integrations times of the DISB registers, using the SetRegister function.
#Set to 10 ms. = 50.000 * 200 ns = 10 ms
SetRegister(ftHandleOutCom, 10, 0)
SetRegister(ftHandleOutCom, 9, 50000)

#Read registers to confirm values were set correctly
IntMSB = ReadRegister(ftHandleOutCom, 10)
print("Integration time IntMSB ",IntMSB)
IntLSB = ReadRegister(ftHandleOutCom, 9)
print("Integration time IntMSB ",IntLSB)


#Get spectrum with integration time 10 ms, using the GetSpectrum function.
Spectrum = GetSpectrum(ftHandleOutCom, ftHandleOutStream)

print("\nPrint spectrum 10ms")
#Print spectrum values
for i in range(len(Spectrum)):
    print(Spectrum[i], end = ' ')



#Set to 100 ms. = 500.000 * 200 ns = 100 ms. Byte 111 1010000100100000 
#MSB 111 = 7
#LSB 10100001 00100000 = 41248
SetRegister(ftHandleOutCom, 10, 7)
SetRegister(ftHandleOutCom, 9, 41248)


#Get spectrum with integration time 100 ms
Spectrum = GetSpectrum(ftHandleOutCom, ftHandleOutStream)

print("\nPrint spectrum 100ms")
#Print spectrum values
for i in range(len(Spectrum)):
    print(Spectrum[i], end = ' ')


##################
#Close devices
##################
ftd2xx.FT_Close(ftHandleOutCom)
ftd2xx.FT_Close(ftHandleOutStream)
ftd2xx.FT_Close(ftHandleOutGPIO)





